#!/usr/bin/python
'''
List current EC2 instances
'''

from bellatrix.lib.ec2_lib import Ec2lib
from bellatrix.lib import bellatrix_util
import bellatrix

import logging

class Lister():
    def __init__(self, key, sec):
        self._ec2 = Ec2lib(key, sec)
        self.space = " | "
        self.attributes = ['architecture', 'block_device_mapping', 'connection', 'dns_name',\
                           'group_name', 'groups', 'hypervisor', 'image_id', 'instance_type',\
                           'ip_address', 'kernel', 'key_name', 'launch_time',\
                           'private_dns_name', 'private_ip_address', 'public_dns_name', 'region',\
                           'root_device_name', 'root_device_type', 'shutdown_state',\
                           'spot_instance_request_id', 'state', 'state_code', 'state_reason',\
                           'subnet_id', 'tags', 'virtualizationType', 'vpc_id']

    def printInstancesInformation(self, instances):
        total = 0
        running = 0
        for i in instances:
            msg = str(i)
            for a in self.attributes:
                try:
                    attr = getattr(i, a)
                    msg += self.space + a + ":" + attr 
                except:
                    logging.debug("Warning. I couldn't find the attribute %s in instance %s" % (a, str(i)))
            print msg
            total += 1
            if i.state == bellatrix.RUNNING_STRING:
                running += 1 
            print ""
        print "Running instances (you pay for them):%s" % running
        print "Total instances:%s" % total

    def list(self):
        logging.info("Getting instances information... (this operation usually takes some seconds)")
        instances = self._ec2.getAllInstances() 
        if len(instances)==0:
            logging.info("There are no EC2 instances in your account.")
        else:
            self.printInstancesInformation(instances)
    

def list():
    l = Lister(bellatrix_util.getKey(), bellatrix_util.getSecret())
    l.list()
    return 0



#!/usr/bin/python
'''
This script runs an ec2 instance and then applies a configuration to it 
'''

import datetime
import logging
import os
import pkgutil
import re
import sys

import bellatrix
from bellatrix.lib import util
from bellatrix.lib import bellatrix_util

from bellatrix.lib.ec2_lib import Ec2lib

class Bewitch():
    def __init__(self, key, sec, app_name, pk, reports): 
        bellatrix_util.checkPkFile(pk)
        self._ec2 = Ec2lib(key, sec) 
        #todo: take a look at: http://stackoverflow.com/questions/1389180/python-automatically-initialize-instance-variables
        self.key = key
        self.sec = sec
        self.pk = pk
        self.app_name = app_name
        self.CMD_OK = 0
        self.define_constants()
        self.reports = reports
        self.out_file = bellatrix_util.getOutFile(__file__)

    def define_constants(self):
        """define class constants to access ami configs"""
        self.AMIS = "amis"
        self.CMDS = "commands"
        self.USER = "user"
        self.BURN_OR_NOT = "burn_ami_at_the_end"
        self.SKIP_ME = "skip_me"
        self.ALL_CONFIGS = "all"
        self.ACCOUNT_LEN = 12
        self.CMD = "command"
        self.CONFIG_DIR = "./configs"  #todo get the path from the script but maybe we will want to deprecate this option
        self.KEY_NAME = "key_name"
        self.SECURITY_GROUPS = "security_groups"
        self.INSTANCE_TYPE = "instance_type"
    
    def getConfigs(self):
        """get configurations from 'configs' directory"""
        #TODO: test this with the new splitted env
        import configs
        cfgpath = os.path.dirname(configs.__file__)
        dir = os.path.basename(cfgpath)
        return [name for _, name, _ in pkgutil.iter_modules([cfgpath])]
        
    def executeCommands(self, user, dns, key, commands, config):
        results = []
        errors = []
        #use the key in the command only if we get one 
        context = {'key': key, 'user':user, 'dns':dns, 'out_tmp': bellatrix.OUT_TMP}
        for c in commands:
            if 'dict' in str(type(c)):
                context['key'] = key
                cmd = c[self.CMD] % context
            else:
                #todo: check if we can capture the stderr and stdout and seeing the output in real time. Maybe using tee?
                context['key'] = ' -i ' + key  
                cmd = "ssh -t -t -o StrictHostKeyChecking=no %(key)s %(user)s@%(dns)s '%(command)s 2>&1' " % \
                dict(context.items() + {self.CMD: c}.items())   #join the two dictionaries
            logging.info("executing: " + cmd)
            res = os.system(cmd)
            out = "" #open(bellatrix.OUT_TMP).read()
            cmd_res = [cmd, out, res, config]
            results.append(cmd_res)
            logging.info("result: " + str(res) + " output: " + out)
            #increment errors if necessary
            if res != 0:
                errors.append(cmd_res)
                break
        logging.info("Commands executions: %s Errors: %s" % (len(commands), len(errors)))
        return results, errors

    def printErrors(self, errors):
        if len(errors) < 1:
            return                
        logging.warning("The following commands failed its execution:")
        for e in errors:
            logging.warning("config: %s cmd: %s exit code: %s" % (e[3], e[0], e[2])) 
            logging.warning("last 500 chars output: %s" % e[1][-500:]) 
            
    def saveReport(self, results, config):                
        logging.info("Saving report")
        report_name = self.reports + os.path.sep + config + "-" + datetime.datetime.now().isoformat() + ".txt"
        report_name_dir = os.path.dirname(report_name)
        if not os.path.isdir(report_name_dir):
            os.makedirs(report_name_dir)
        with open(report_name, "w") as f:
            for r in results:
                f.write("res: %s cmd: %s out: %s \n" % (r[2], r[0], r[1]))
        logging.info("Report saved in: %s" % report_name) 

    def _processConfig(self, amis, commands, user, burn_at_the_end, key_name, security_groups, instance_type):
        """execute a configuration, internal method of run"""
        amis_burned = []
        errors = []
        for ami in amis:
            config_name = ami[1]
            ami = ami[0]
            inst, dns_name = self._ec2.startInstance(ami, instance_type, key_name, security_groups)
            instance_name = inst.id
            self._ec2.waitForConnectionReady(inst, user, self.pk, dns_name)
            r, e = self.executeCommands(user, inst.dns_name, self.pk, commands, config_name)
            self.saveReport(r, config_name)
            errors += e
            if len(e) > 0:
                logging.warning("There were errors while executing the commands. Not burning the instance...")
            else:
                if burn_at_the_end:
                    try:
                        new_ami = self._ec2.createImage(inst.id, config_name + "-" 
                                                        + datetime.datetime.now().isoformat(), "generated by " 
                                                        + self.app_name)
                        logging.info("ami: %s is being generated for configuration: %s" 
                                     % (new_ami, config_name))
                        amis_burned.append([new_ami, config_name])
                        util.writeFile(self.out_file, new_ami + "," + instance_name + os.linesep)
                    except:
                        logging.exception("Problem burning image: %s with instance: %s" % (config_name, inst.id))
            return amis_burned, errors

    def run(self, config):
        """execute a configuration"""
        amis_burned = []
        errors = []
        if config == self.ALL_CONFIGS:
            configs = self.getConfigs()
        else:
            configs = [os.path.splitext(config)[0]]
        sys.path = [util.getCurDir()] + sys.path
        for cfg in configs:
            logging.info("processing: " + cfg + " in: " + os.getcwd())
            if config == self.ALL_CONFIGS:
                c = __import__(os.path.basename(self.CONFIG_DIR) + "." + cfg)
                mod = "c." + cfg + "."
            else:
                c = __import__(cfg)
                mod = "c."
            skip_me = eval(mod + self.SKIP_ME)
            if skip_me:
                logging.info("skipping execution of config: %s due to its configuration skip_me=true" % cfg)
                continue
            amis = eval(mod + self.AMIS)
            commands =  eval(mod + self.CMDS)
            user =  eval(mod + self.USER)
            burn_at_the_end = eval(mod + self.BURN_OR_NOT)
            a,e = self._processConfig(amis, commands, user, burn_at_the_end, eval(mod + self.KEY_NAME), 
                                      eval(mod + self.SECURITY_GROUPS), eval(mod + self.INSTANCE_TYPE))
            amis_burned += a
            errors += e
        self.printErrors(errors)
        logging.info("total of ami's burned:%s" % len(amis_burned))
        for a in amis_burned:
            logging.info(str(a))
        return 0 if len(errors)==0 else 1

    
def run(configuration=None):
    r = Bewitch(bellatrix_util.getKey(), bellatrix_util.getSecret(), bellatrix.APP, 
            bellatrix_util.getPrivateKey(), bellatrix_util.getReportsDir())
    config = r.ALL_CONFIGS if (not configuration) else configuration
    exit_code = r.run(config)
    return exit_code

if __name__ == '__main__':
    sys.exit(run(sys.argv))




#!/usr/bin/python
'''
Modify security groups
'''

import sys

from bellatrix.lib.ec2_lib import Ec2lib
from bellatrix.lib import util
from bellatrix.lib import bellatrix_util



class Run():
    def __init__(self, key, sec):
        self._ec2 = Ec2lib(key, sec) 

    def setSecurityGroupAuth(self, name, ports, cidrs):
        ports = util.getStringsFromFile(ports)
        cidrs = util.getStringsFromFile(cidrs)
        sg = self._ec2.getSecurityGroups(name)[0]
        self._ec2.revokeAllSecurityGroupRules(sg)
        for p in ports:
            for c in cidrs:
                self._ec2.authorizeSecurityGroup(sg, c, p)
            #allow same instances connection
            self._ec2.authorizeSecurityGroup(sg, None, p, sg)
                     

def run(security_group_name, ports_file, cidrs_file):
    r = Run(bellatrix_util.getKey(), bellatrix_util.getSecret())
    r.setSecurityGroupAuth(security_group_name, ports_file, cidrs_file)
    return 0

    

if __name__ == '__main__':
    sys.exit(run(*sys.argv[1:]))






#Constants

APP = "Bellatrix"
__version__ = '1.1.8'
description = "AWS EC2 magic utilities"

SECRET_FILE = "secret"
KEY_FILE = "key"
#path to the private key to connect to agents
PRIVATE_KEY_FILE = "ec2.pk"

OUT_TMP = "exec.tmp"


#---------------------------------------------------
#AWS constants

#shutdown behaviour
TERMINATE = "terminate"  
AMI_AVAILABLE = "available"
#running state for an ec2 instance
RUNNING = 16    
RUNNING_STRING = "running"



""" 
Common utilities, meant to be used as 'building blocks'.
Every block has a different use, being possible in some cases to composite them. 

Every command is a simple function that returns the list of commands to be executed 
into a list of strings.
If any of the commands fails, the execution will be interrupted. 
It's convenient that you add a small test at the very end of each set. 

For example:

def install_pip():
    return [
            "touch new_file",
            "cat new_file",
            ] 
               
It's an example of a single-purposed operation that performs a test at the end. 
Should something happened to the operation, 'cat new_file' will fail. 
Additionally you will be able to verify the content of the file in the logging output.               
"""

import os

#CONSTANTS
#http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=422427
apt_prefix = "export PATH=$PATH:/sbin:/usr/sbin:/usr/local/sbin;export TERM=linux; "
apt_get_and_options = "sudo /usr/bin/apt-get -q -y -o DPkg::Options::=--force-confold --fix-missing " 
apt_install = apt_get_and_options + "install "
apt_get_install_cmd = apt_prefix + apt_install
cmd = "command" 


def apt_get_install(package):
    """
    Return the "sudo apt-get install" command
    
    :type package: string
    :param package: The new package to install
    """
    return [apt_get_install_cmd + package]

def pip_install(package, prefix="sudo", verify=True, verification_command=None):
    """
    Install a Python package using pip.
    
    :type package: string
    :param package: The package to install.    

    :type prefix: string
    :param prefix: prefix of the pip command. It is 'sudo' by default.     

    :type verify: bool
    :param verify: True by default. If true, it will auto-generate a verification command to check the package has been installed correctly.      

    :type verification_command: string
    :param verification_command: None by default. If None, the verification command will be 'package --version' Otherwise it will execute the one provided in this argument.      
    """
    cmds = [prefix + " pip install " + package + " --upgrade"]
    if verification_command:
        cmds.append(package + " --version" if verification_command == None else verification_command)
    return cmds

def createVirtualEnv(env_name):
    """
    Generate a new Python virtual environment using virtualenv
    
    :type env_name: string
    :param env_name: Name of the new virtual environment.
    """
    return ["virtualenv --no-site-packages --clear " + env_name]

def executeInVirtualEnv(env, cmd):
    """
    Execute a command within a virtualenv environment
    
    :type env: str
    :param env: Name of the new virtual environment.

    :type cmd : list
    :param cmd: List of commands to be executed.
    """
    if type(cmd) == type(list()):
        cmd = flatCommands(cmd) 
    return ["source " + env + os.path.sep + "bin" + os.path.sep + "activate && " + cmd] 

def flatCommands(cmds):
    """
    Given a list of commands it will return a single one concatenated by '&&' so they will be executed in sequence until any of them fails.
    
    :type cmds: list
    :param cmds: List of strings that contains commands.
    
    :rtype: string
    :return: A single string with the commands concatenated.
    """
    cmd = ""
    ampersand = " && "
    #we need to get a single command so it works in the virtual environment
    for c in cmds:
        cmd += c + ampersand 
    cmd = cmd[:len(cmd) - len(ampersand)]   #cut last &&
    return cmd


def installPackageInVirtualEnv(env, package, verify=True, verification_command=None, prefix=""):
    """
    Install a Python package into a virtualenv.
    
    :type env: string
    :param env: Name of the virtual environment.

    :type package: string
    :param package: Package to be installed.

    :type verify: string
    :param verify: True by default. If true, it will auto-generate a verification command to check the package has been installed correctly.      

    :type verification_command: string
    :param verification_command: None by default. If None, the verification command will be 'package --version' Otherwise it will execute the one provided in this argument.      

    :type prefix: string
    :param prefix: Prefix of the pip command. It is 'sudo' by default.     
    """
    cmds = pip_install(package, prefix, verify, verification_command) 
    cmd = flatCommands(cmds)
    return executeInVirtualEnv(env, cmd) 

def install_pip():
    """
    Install pip using apt-get install.
    """
    #we need to upgrade pip package using pip itself
    return apt_get_install("python-pip") + pip_install("pip")    
               
def apt_get_update():
    """
    Executes apt-get update.
    """
    return [
            apt_get_and_options + "update"
            ]

def install_nginx():
    """
    Install Nginx in Ubuntu using the repo they provide as described in http://wiki.nginx.org/Install#Ubuntu_PPA
    """
    return ["sudo add-apt-repository ppa:nginx/stable -y"] + apt_get_update() + apt_get_install("nginx")
                
def create_django_project(project_name, dir_name="." + os.path.sep):                
    """
    Creates a Django project.
    
    :type project_name: string
    :param project_name: Name of the Django project.

    :type dir_name: string
    :param dir_name: Directory where the Django project needs to be executed. By default is the current directory.  
    """
    return ["cd " + dir_name + " && rm -rf " + project_name + " && django-admin.py startproject " + project_name]

def wget(url, dest=None):
    """
    Downloads a web resource using wget.
    
    :type url: string
    :param url: Url to download.

    :type dest: string
    :param dest: Optional. Destination in our machine of the downloaded resource.
    """
    return ["wget --no-check-certificate " + url + ("" if dest == None else " -O " + dest)]

def sudo(cmds):
    """
    Execute a list of commands using sudo.
    
    :type cmds: list
    :param cmds: List of commands.
    """
    for i in range(len(cmds)):
        cmds[i] = "sudo " + cmds[i]
    return cmds

def chmod(mode, file_name, options=""):
    """
    Applies the chmod command.
    
    :type mode: string
    :param mode: New mode of the destination. Typical examples are: a+r to give read permissions to everyone, a+rx to give execution and read permissions to any user, etc.

    :type file_name: string
    :param file_name: Destination of the new mode.

    :type options: string
    :param options: Optional set of parameters for this command.   
    """
    return ["chmod " + options + " " + mode + " " + file_name]    
    
def mkdir(directory):
    """
    Created a new directory. "-p" flag is used so the command generates the same result regardless whether the directory exists or not. 
    
    :type directory: string
    :param directory: directory to be created.
    """
    return ["mkdir -p " + directory]    

def createSoftLink(src, dest):
    """
    Creates a new soft link.
    
    :type src: string
    :param Src: Source file or directory.

    :type dest: string
    :param Dest: Name of the soft link.
    """
    return ["ln -f -s %s %s" % (src, dest)]    

def copy(src, dest):
    """
    Copy a file using -f so it doesn't fail if the destination exists.

    :type src: string
    :param src: Source file.

    :type dest: string
    :param dest: File destination.
    """
    return ["cp -f %s %s" % (src, dest)]    

def scp(src, dest):
    return [
            "mkdir -p %s" % dest,
            {cmd:"scp -o StrictHostKeyChecking=no -i %(key)s " 
             + src + " %(user)s@%(dns)s:" 
             + dest} 
            ]

def logical_resize_disk(device):
    """
    Resizes the logical volume to new_size specified in the start command.  This is necessary
    for some operating systems that do not automatically resize to this value.
    
    :type device: string
    :param device: logical volume to resize, e.g. /dev/xvda1 in RedHat family systems 

    """
    return ["resize2fs " + device]



'''
general utility functions
'''

import logging
import os
import time
import xml.dom.minidom


def getNodeValue(dom, nodename):
    """
    get 'myname' from a xml structure like this one when you call this method with 'name':
    <?xml version="1.0" encoding="UTF-8" ?>
    <root>
    <name>myname</name>
    </root>
    """
    return dom.getElementsByTagName(nodename).item(0).firstChild.nodeValue

def getDomFromXmlFile(xml_file):
    """
    parse an xml file and return its dom 
    """
    return xml.dom.minidom.parse(xml_file) 

def getValue(xml_file, value):
    """
    return the value from a xml tag

    this function will get 'myname' from a xml structure like this one when you call this method with 'name':
    <?xml version="1.0" encoding="UTF-8" ?>
    <root>
    <name>myname</name>
    </root>
    """
    return getNodeValue(getDomFromXmlFile(xml_file), value)

def getConfigDir(app_name):
    config_dir = None
    return config_dir
    
def getConfigFileContent(file_name, app_name):
    content = None
    return content

def getHome():
    return os.path.expanduser('~') or os.getenv('USERPROFILE') or os.getenv('HOME')

def getCurDir():
    return os.path.abspath(os.getcwd())

def getStringsFromFile(list_file):
    """"Return list from file ignoring blanks and comments"""
    l = []
    with open(list_file) as f:
        for line in f:
            line = line.strip()
            if len(line) > 0 and not line.startswith("#"):
                l.append(line)
    return l

def writeFile(file_name, content):
    with open(file_name, 'w') as out:
        out.write(content)

def importModule(module):
    dir_name = os.path.dirname(os.path.abspath(module))
    if len(dir_name) > 0 and os.path.sep in dir_name:
        #we change a normal path to a python module
        module = module.replace(os.path.sep, ".")
    return __import__(os.path.basename(module))


def waitForSSHReady(user, key, dns, TIME_OUT=300):
    ERR_CONNECTION_REFUSED = 65280 
    logging.info("waiting until host: " + dns + " is ready to receive ssh connections. Time out is: " + str(TIME_OUT) + " seconds...")
    tmp_file = "tmp"
    cmd = "ssh -o StrictHostKeyChecking=no -i %s %s@%s '%s' > %s" % (key, user, dns, "echo CONNECTION READY", tmp_file)
    step = 3
    result = ERR_CONNECTION_REFUSED
    while (result == ERR_CONNECTION_REFUSED and TIME_OUT > 0):
        TIME_OUT -= step
        time.sleep(step)
        logging.info("executing:%s " % cmd )
        result = os.system(cmd)
    if result !=  0:
        raise Exception("Sorry, but the instance never got ready for SSH connections")
    logging.info("Host %s is ready to receive ssh connections." % (dns))



#!/usr/bin/python
'''
get a new ami given a running instance 
'''

import datetime
import logging
import sys
import time

import bellatrix
from bellatrix.lib.ec2_lib import Ec2lib
from bellatrix.lib import bellatrix_util
from bellatrix.lib import util


class Run():
    def __init__(self, key, sec, app_name):
        self.WAIT = 30 
        self._ec2 = Ec2lib(key, sec) 
        self._app_name = app_name
        self.out_file = bellatrix_util.getOutFile(__file__)
        

    def burnInstance(self, instance, config_name):
        config_name = config_name + "-" + datetime.datetime.now().isoformat()
        new_ami = self._ec2.createImage(instance, config_name, "generated by " 
                                        + self._app_name)
        logging.info("ami: %s is being generated. Name: %s" 
                     % (new_ami, config_name))
        util.writeFile(self.out_file, new_ami + "," + config_name)
        return new_ami
    
    def amiIsReady(self, ami_name):
        ready = False
        exceptions = 5
        while not ready:
            try:
                ami = self._ec2.getImage(ami_name)
                logging.info("ami state %s" % ami.state)
                ready = (ami.state == bellatrix.AMI_AVAILABLE)
            except:
                #we will swallow the exception a number of times 
                #since the image object might not be ready at the beginning
                exceptions -= 1
                if exceptions < 1:
                    raise Exception("""Error getting information from image: %s. """
                                    """Please check if you have access or if the image name is correct.""" 
                                    % ami_name)
            if not ready:
                logging.info("AMI:%s is not ready yet. Retrying in %s seconds.... " % (ami_name, self.WAIT))
                time.sleep(self.WAIT)

        logging.info("AMI: %s is ready to use!" % ami_name)

def run(instance, config_name, wait):
    r = Run(bellatrix_util.getKey(), bellatrix_util.getSecret(), bellatrix.APP)
    new_ami = r.burnInstance(instance, config_name)
    if wait:
        r.amiIsReady(new_ami)
    return 0

if __name__ == '__main__':
    sys.exit(run(*sys.argv[1:])) 




#!/usr/bin/python
'''
Set launch pemissions for an AMI
'''

import logging
import sys

from bellatrix.lib.ec2_lib import Ec2lib
from bellatrix.lib.bellatrix_util import *


class Run():
    def __init__(self, key, sec): 
        self._ec2 = Ec2lib(key, sec) 
        self.define_constants()

    def define_constants(self):
        """define class constants to access ami configs"""
        self.ACCOUNT_LEN = 12
    
    def getAccountPermissions(self, perm_file):
        """"Return list of accounts where new ami's will get execute permissions"""
        l = []
        with open(perm_file) as f:
             for line in f:
                 line=line.strip()
                 if len(line)==self.ACCOUNT_LEN:
                     l.append(line)
        logging.info("accounts from %s: %s" % (perm_file,l))
        return l
    
    def setPermissions(self, amis, permissions):
        self._ec2.setPermissionsToAmis(amis, permissions)
                     

def run(ami, permissions_file):
    r = Run(getKey(), getSecret())
    r.setPermissions(ami.split(","), r.getAccountPermissions(permissions_file))
    return 0


if __name__ == '__main__':
    sys.exit(run(*sys.argv[1]))
